// psterm.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

BOOL CheckMembership(HANDLE hToken, PSID pSid, PBOOL pbIsMember)
{
    BOOL bClose = FALSE,
         bResult = FALSE;

    if (!hToken)
    {
        HANDLE hThread = NULL;

        OpenThreadToken(GetCurrentThread(), TOKEN_QUERY, FALSE, &hToken);
        bClose = TRUE;

        if (hThread)
            CloseHandle(hThread);
    }

    if (!hToken)
        return bResult;

    BYTE dacl[sizeof(ACL) + (sizeof(ACCESS_ALLOWED_ACE) << 1)];

    InitializeAcl((PACL) &dacl[0], sizeof(dacl), ACL_REVISION);
    AddAccessAllowedAce((PACL) &dacl[0], ACL_REVISION, 1, pSid);
    
    SECURITY_DESCRIPTOR sd;
    InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION);
    SID world = {SID_REVISION, 1, SECURITY_WORLD_SID_AUTHORITY, SECURITY_WORLD_RID};

    SetSecurityDescriptorOwner(&sd, &world, FALSE);
    SetSecurityDescriptorGroup(&sd, &world, FALSE);
    SetSecurityDescriptorDacl(&sd, TRUE, (PACL) &dacl[0], FALSE);

    GENERIC_MAPPING gm = {0,0,0,1};
    PRIVILEGE_SET ps;
    DWORD cb = sizeof(ps), ga;

    bResult = AccessCheck(&sd, hToken, 1, &gm, &ps, &cb, &ga, pbIsMember);

    if (bClose)
        CloseHandle(hToken);

    return bResult;
}

BOOL IsUserAdmin(VOID)
{
    BOOL bResult;
    SID_IDENTIFIER_AUTHORITY NtAuthority = SECURITY_NT_AUTHORITY;
    PSID AdministratorsGroup; 
    bResult = AllocateAndInitializeSid(
                                        &NtAuthority,
                                        2,
                                        SECURITY_BUILTIN_DOMAIN_RID,
                                        DOMAIN_ALIAS_RID_ADMINS,
                                        0, 0, 0, 0, 0, 0,
                                        &AdministratorsGroup
                                      ); 
    if(bResult) 
    {
        if (!CheckMembership(NULL, AdministratorsGroup, &bResult)) 
        {
             bResult = FALSE;
        } 
        FreeSid(AdministratorsGroup); 
    }

    return bResult;
}


VOID PrintError(const char *pcszPrefix, DWORD dwError, HMODULE hModule)
{
	LPVOID	lpMsgBuf = NULL;
    DWORD dwChars = 0;
 
    if (hModule)
    {
	    dwChars = FormatMessage(
                        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_HMODULE,
					    hModule, 
                        dwError,
					    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                        (LPTSTR) &lpMsgBuf, 
                        0, 
                        NULL 
                     );
    }

    if (!dwChars)
    {
	    dwChars = FormatMessage(
                        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
					    NULL, 
                        dwError,
					    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                        (LPTSTR) &lpMsgBuf, 
                        0, 
                        NULL 
                     );
    }

    if (dwChars)
    {
	    fprintf(stderr, "%s%s", pcszPrefix, lpMsgBuf);
    }
    else
        fprintf(stderr, "%s0x%08x\n", pcszPrefix, dwError);

    if (lpMsgBuf)
        LocalFree(lpMsgBuf);

}

int main(int argc, char* argv[])
{
    int iResult = ~0;

    if (argc != 2)
    {
        fprintf(stderr, "PSTERM: USAGE - psterm <ImageName | PID>\n");
        return iResult;
    }

    // Administrators only
    ImpersonateSelf(SecurityImpersonation);

    if (!IsUserAdmin())
    {
        RevertToSelf();
        SetLastError(iResult = ERROR_ACCESS_DENIED);
        PrintError("PSTERM: ", GetLastError(), NULL);
        return iResult;
    }

    RevertToSelf();

    DWORD dwBytes = 0, dwPID = strtoul(argv[1], NULL, 0);
	HANDLE hDevice = CreateFile("\\\\.\\PROCMAP", GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);

    if (hDevice != INVALID_HANDLE_VALUE)
    {
        // Try by PID first
        if (dwPID && DeviceIoControl(hDevice, 0x8000a000, &dwPID, sizeof(dwPID), NULL, 0, &dwBytes, NULL))
        {
            fprintf(stdout, "PSTERM: Process %u terminated.\n", dwPID);
            iResult = 0;
        }
        else
        {
           // Try by image name
           if (DeviceIoControl(hDevice, 0x8000a004, argv[1], strlen(argv[1]), NULL, 0, &dwBytes, NULL))
           {
                fprintf(stdout, "PSTERM: Process %s terminated.\n", argv[1]);
                iResult = 0;
           }
           else
           {
                PrintError("PSTERM: ", iResult = GetLastError(), NULL);
           }
        }
    }
    else
    {
        PrintError("PSTERM: ", iResult = GetLastError(), NULL);
    }

    if (hDevice != INVALID_HANDLE_VALUE)
        CloseHandle(hDevice);

	return iResult;
}

